/***************************************************************************
 *
 * gpn.c : GNOME Power Notification 
 *         (Panel functions for GNOME Power Manager)
 *
 * Copyright (C) 2005 Richard Hughes, <richard@hughsie.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 **************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gconf/gconf-client.h>
#include <gdk/gdk.h>
#include "common.h"
#include "icons.h"
#include "gpm.h"
#include "gpn.h"

/* shared with gpm.c */
HasData has_data;
StateData state_data;
static IconData icon_main, icon_low;
GPtrArray *objectData;

/** Calculate the color of the charge level bar
 *
 *  @param  level		the chargeLevel of the object, 0..100
 *  @return			the RGB colour
 */
static guint32
level_bar_get_color (guint level)
{
	guint8 red, green;
	gfloat c;

	red = green = 204; /* 0xCC */
	if (level < 50) {
		c = ((gfloat) level) / 50.0;
		green = (guint8) (gfloat) (204.0 * c);
	}
	if (level > 50) {
		c = (100.0 - (gfloat) level) / 50.0;
		red = (guint8) (gfloat) (204.0 * c);
	}
	return (0xff + red * 0x01000000 + green * 0x00010000);
}

static GdkPixbuf *
gtk_icon_theme_fallback (const char *name, int size)
{
	GdkPixbuf *pixbuf = NULL;
	GError *err = NULL;
	g_debug ("name = '%s'\n", name);
	if (gtk_icon_theme_has_icon (gtk_icon_theme_get_default (), name)) {
		GtkIconInfo *iinfo = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (), name, size, GTK_ICON_LOOKUP_USE_BUILTIN);
		pixbuf = gtk_icon_info_load_icon (iinfo, &err);
		gtk_icon_info_free (iinfo);
	} else {
		/* 
		 * We cannot find this specific themed GNOME icon so use builtin
		 * fallbacks. This makes GPM more portible between distros
		 */
		GString *fallback = g_string_new ("error?");
		g_string_printf (fallback, "%s%s.png", GPM_DATA, name);
		pixbuf = gdk_pixbuf_new_from_file (fallback->str, &err);
	}
	return pixbuf;
}

static GdkPixbuf *
create_icon_pixbuf (GenericObject *object)
{
	g_assert (object);
	GError *err = NULL;
	GdkPixbuf *pixbuf = NULL;
	GString *name = convert_powerdevice_to_gnomeicon (object->powerDevice);
	g_assert (name);
	pixbuf = gtk_icon_theme_fallback (name->str, 24);
	g_assert (pixbuf);
	g_string_free (name, TRUE);

	/* merge with AC emblem if necessary */
	if (object->isCharging) {
		GdkPixbuf *emblem = gdk_pixbuf_new_from_file (GPM_DATA "emblem-ac.png", &err);
		g_assert (emblem);
		gdk_pixbuf_composite (emblem, pixbuf, 0, 0, 24, 24, 0, 0, 1.0, 1.0, GDK_INTERP_BILINEAR, 0xFF);
		g_object_unref (emblem);
	}

	/* merge with level bar if necessary */
	if (object->isRechargeable) {
		GdkPixbuf *emblem = gdk_pixbuf_new_from_file (GPM_DATA "emblem-bar.png", &err);
		g_assert (emblem);
		gdk_pixbuf_composite (emblem, pixbuf, 0, 0, 24, 24, 0, 0, 1.0, 1.0, GDK_INTERP_BILINEAR, 0xFF);
		g_object_unref (emblem);

		gfloat c = ((gfloat) object->percentageCharge) / 100.0;
		guint h = (guint) (19.0 * c);
		if (h < 1)h = 1;
		if (h > 19) h = 19;

		GdkPixbuf *bar = gdk_pixbuf_new_subpixbuf (pixbuf, 20, 22 - h, 3, h);
		gdk_pixbuf_fill (bar, level_bar_get_color (object->percentageCharge));
		g_object_unref (bar);
	}
	return pixbuf;
}

/* wrapper function */
void
gpn_icon_initialise (gint isMain)
{
	IconData *icon;
	GConfClient *client = gconf_client_get_default ();

	if (isMain & ICON_LOW) {
		icon = &icon_low;
		free_icon_structure (icon);
		icon->isMain = FALSE;
		icon->show = gconf_client_get_bool (client, "/apps/gnome-power/general/displayIconLow", NULL);
		icon->displayOptions = gconf_client_get_int (client, "/apps/gnome-power/general/lowThreshold", NULL);
	}
	if (isMain & ICON_MAIN) {
		icon = &icon_main;
		free_icon_structure (icon);
		icon->isMain = TRUE;
		icon->show = gconf_client_get_bool (client, "/apps/gnome-power/general/displayIcon", NULL);
	}
}

/* wrapper function */
void
gpn_icon_destroy (gint isMain)
{
	IconData *icon;
	if (isMain & ICON_LOW) {
		icon = &icon_low;
		if (icon->td)
			icon_destroy (&(icon->td));
		free_icon_structure (icon);
	}
	if (isMain & ICON_MAIN) {
		icon = &icon_main;
		if (icon->td)
			icon_destroy (&(icon->td));
		free_icon_structure (icon);
	}
}

/* wrapper function */
void
gpn_icon_update (gint isMain)
{
	if (isMain & ICON_MAIN)
		update_icon (&icon_main);
	if (isMain & ICON_LOW)
		update_icon (&icon_low);
}

/* wrapper function */
void
callback_gconf_key_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
	g_assert (client);
	g_assert (entry);
	g_debug ("callback_gconf_key_changed (%s)", entry->key);

	if (gconf_entry_get_value (entry) == NULL)
		return;

	if (strcmp (entry->key, "/apps/gnome-power/general/displayIcon") == 0) {
		icon_main.show = gconf_client_get_bool (client, entry->key, NULL);
		gpn_icon_update (ICON_MAIN);
	} else if (strcmp (entry->key, "/apps/gnome-power/general/displayIconLow") == 0) {
		icon_low.show = gconf_client_get_bool (client, entry->key, NULL);
		gpn_icon_update (ICON_LOW);
	} else if (strcmp (entry->key, "/apps/gnome-power/general/lowThreshold") == 0) {
		icon_low.displayOptions = gconf_client_get_int (client, entry->key, NULL);
		gpn_icon_update (ICON_LOW);
	}
}

/** Gets the timestring from a slot object
 *
 *  @param  slotData		the GenericObject reference
 *  @return			the timestring, e.g. "13 minutes until charged"
 */
static GString *
get_time_string (GenericObject *slotData)
{
	g_assert (slotData);
	g_debug ("get_time_string");
	GString* timestring = NULL;
	timestring = get_timestring_from_minutes (slotData->minutesRemaining);
	if (!timestring)
		return NULL;
	if (slotData->isCharging)
		timestring = g_string_append (timestring, " until charged");
	else
		timestring = g_string_append (timestring, " remaining");

	return timestring;
}

/** Returns the tooltip for the main icon. Text logic goes here :-)
 *
 */
GString *
get_main_tooltip (IconData *tdicon)
{
	g_assert (tdicon);
	g_assert (tdicon->slotData);
	g_debug ("get_main_tooltip");
	GenericObject *slotData = NULL;
	GString *chargestate = NULL;
	GString *remaining = NULL;
	GString *devicestr = NULL;
	GString *tooltip = NULL;
	gint a;

	if (state_data.onBatteryPower)
		tooltip = g_string_new ("Computer is running on battery power");
	else if (state_data.onUPSPower)
		tooltip = g_string_new ("Computer is running on UPS power");
	else
		tooltip = g_string_new ("Computer is running on AC power");

	for (a=0;a<objectData->len;a++) {
		slotData = (GenericObject *) g_ptr_array_index (objectData, a);
		devicestr = convert_powerdevice_to_string (slotData->powerDevice);
		if (slotData->powerDevice == POWER_PRIMARY_BATTERY ||
		    slotData->powerDevice == POWER_UPS) {
			remaining = get_time_string (slotData);
			chargestate = get_chargestate_string (slotData);
			g_string_append_printf (tooltip, "\n%s %s (%i%%)", 
						devicestr->str, chargestate->str, slotData->percentageCharge);
			if (remaining)
				g_string_append_printf (tooltip, "\n%s", remaining->str);
		} else if (slotData->powerDevice == POWER_KEYBOARD ||
			   slotData->powerDevice == POWER_MOUSE) {
			g_string_append_printf (tooltip, "\n%s (%i%%)", 
						devicestr->str, slotData->percentageCharge);
		}
		if (devicestr)
			g_string_free (devicestr, TRUE);
		if (chargestate)
			g_string_free (chargestate, TRUE);
		if (remaining)
			g_string_free (remaining, TRUE);
		devicestr = NULL;
		chargestate = NULL;
		remaining = NULL;
	}
	return tooltip;
}

/** Returns the tooltip for the main icon. Text logic goes here :-)
 *
 */
GString *
get_low_tooltip (IconData *tdicon)
{
	g_assert (tdicon);
	g_assert (tdicon->slotData);
	g_debug ("get_low_tooltip");
	GString *tooltip;
	tooltip = convert_powerdevice_to_string (tdicon->slotData->powerDevice);
	g_string_append_printf (tooltip, " is very low! (%i%%)", tdicon->slotData->percentageCharge);
	return tooltip;
}

/** Finds the best selection for the "main" icon in the notification area
 *
 *  @return  		The pointer to the main icon, or NULL if none needed
 */
GenericObject *
get_main_icon_slot (void)
{
	GenericObject *slotData;
	gint powerDevice;
	gint a;

	/* get prefered icon */
	if (has_data.hasBatteries)
		powerDevice = POWER_PRIMARY_BATTERY;
	else if (has_data.hasUPS)
		powerDevice = POWER_UPS;
	else
		powerDevice = POWER_AC_ADAPTER;

	/* find an example of that icon */
	for (a=0;a<objectData->len;a++) {
		slotData = (GenericObject *) g_ptr_array_index (objectData, a);
		if (slotData->powerDevice == powerDevice) {
			return slotData;
		}
	}
	return NULL;
}

/** Finds the best selection for the "low" icon in the notification area
 *
 *  @return  		The pointer to the low icon, or NULL if no selection
 *  @todo  		Improve logic so that "lowest" devices are chosen
 */
GenericObject *
get_low_icon_slot (void)
{
	gint a;
	GenericObject *slotData;
	for (a=0;a<objectData->len;a++) {
		slotData = (GenericObject *) g_ptr_array_index (objectData, a);
		if (slotData->powerDevice != POWER_NONE && 
		    slotData->powerDevice != POWER_AC_ADAPTER && 
		    slotData != icon_main.slotData && 
		    slotData->present && 
		    slotData->percentageCharge < icon_low.displayOptions) {
			return slotData;
		}
	}
	return NULL;
}

/** Frees icon structure
 *
 *  @param  tdicon		The icon pointer that needs to be freed
 */
void
free_icon_structure (IconData *tdicon)
{
	g_assert (tdicon);
	g_debug ("free_icon_structure");
	if (tdicon->tooltip)
		g_string_free (tdicon->tooltip, TRUE);
	tdicon->tooltip = NULL;
	tdicon->td = NULL;
}

/** Callback for actions boxes
 *
 */
#if GPM_HAS_ACTIONS
static void
callback_actions_activated (GtkMenuItem *menuitem, gpointer user_data)
{
	char *action = g_object_get_data ((GObject*) menuitem, "action");
	g_debug ("action = '%s'", action);
	if (strcmp (action, "shutdown") == 0)
		action_policy_do (ACTION_SHUTDOWN);
	else if (strcmp (action, "reboot") == 0)
		action_policy_do (ACTION_REBOOT);
	else if (strcmp (action, "suspend") == 0)
		action_policy_do (ACTION_SUSPEND);
	else if (strcmp (action, "hibernate") == 0)
		action_policy_do (ACTION_HIBERNATE);
	else
		g_warning ("actions_activated : No handler for '%s'", action);
}
#endif

/** Callback for "about" box
 *
 */
static void
callback_about_activated (GtkMenuItem *menuitem, gpointer user_data)
{
	g_debug ("callback_about_activated");
	static GtkWidget *about = NULL;
	GdkPixbuf *pixbuf = NULL;
	const gchar *authors[] = { "Richard Hughes <richard@hughsie.com>", NULL };
	const gchar *documenters[] = { NULL };

	if (about)
	{
		gdk_window_raise (about->window);
		gdk_window_show (about->window);
		return;
	}

	pixbuf = gdk_pixbuf_new_from_file (GPM_DATA "battery-48.png", NULL);
	about = gnome_about_new(NICENAME, VERSION,
			"Copyright \xc2\xa9 2005 Richard Hughes",
			_(NICEDESC),
			(const char **)authors,
			(const char **)documenters,
			NULL,
			pixbuf);

	if (pixbuf)
		gdk_pixbuf_unref (pixbuf);

	g_signal_connect (G_OBJECT (about), "destroy", G_CALLBACK (gtk_widget_destroyed), &about);
	g_object_add_weak_pointer (G_OBJECT (about), (void**)&(about));
	gtk_widget_show(about);
}

#if GPM_HAS_QUIT
/** Callback for quit
 *
 */
static void
callback_quit_activated (GtkMenuItem *menuitem, gpointer user_date)
{
	g_debug ("callback_quit_activated");
	GnomeClient *master;
	GnomeClientFlags flags;
	master = gnome_master_client ();
	flags = gnome_client_get_flags (master);
	if (flags & GNOME_CLIENT_IS_CONNECTED) {
		gnome_client_set_restart_style (master,	GNOME_RESTART_NEVER);
		gnome_client_flush (master);
	}
	gpn_icon_destroy (ICON_ALL);
	exit (0);
}
#endif

/** Callback for preferences
 *
 */
static void
callback_prefs_activated (GtkMenuItem *menuitem, gpointer user_data)
{
	g_debug ("callback_prefs_activated");
	gboolean retval;
	gchar *path;

	path = g_strconcat (BINDIR, "/", "gnome-power-preferences", NULL);
	retval = g_spawn_command_line_async (path, NULL);
	if (retval == FALSE)
		g_warning ("Couldn't execute command: %s", path);
	g_free (path);
}

static void
menu_add_separator_item (GtkWidget *menu)
{
	GtkWidget *item;
	item = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	gtk_widget_show (item);
}

#if GPM_HAS_ACTIONS
static void
menu_add_action_item (GtkWidget *menu, const char *icon, const char *name, char *type)
{
	/* get image */
	GtkWidget *image = gtk_image_new ();
	GdkPixbuf *pixbuf = gtk_icon_theme_fallback (icon, 16);
	g_assert (pixbuf);
	gtk_image_set_from_pixbuf (GTK_IMAGE (image), pixbuf);

	GtkWidget *item = gtk_image_menu_item_new_with_label (name);
	g_object_set_data ((GObject*) item, "action", (gpointer) type);
	gtk_image_menu_item_set_image ((GtkImageMenuItem*) item, GTK_WIDGET (image));
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (callback_actions_activated), (gpointer) menu);
	gtk_widget_show (item);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);	
}
#endif

/** Creates panel menu
 *
 *  @return			MenuData object
 */
static void
menu_main_create (TrayData *td)
{
	g_assert (td);
	g_assert (td->popup_menu == NULL);
	g_debug ("menu_main_create");
	GtkWidget *item;
	td->popup_menu = gtk_menu_new ();

	item = gtk_image_menu_item_new_from_stock (GTK_STOCK_PREFERENCES, NULL);
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (callback_prefs_activated), (gpointer) td->popup_menu);
	gtk_widget_show (item);
	gtk_menu_shell_append (GTK_MENU_SHELL (td->popup_menu), item);

	item = gtk_image_menu_item_new_from_stock (GNOME_STOCK_ABOUT, NULL);
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (callback_about_activated), (gpointer) td->popup_menu);
	gtk_widget_show (item);
	gtk_menu_shell_append (GTK_MENU_SHELL (td->popup_menu), item);

#if GPM_HAS_ACTIONS
	menu_add_separator_item (td->popup_menu);
	menu_add_action_item (td->popup_menu, "gnome-reboot", "Reboot", "reboot");
	menu_add_action_item (td->popup_menu, "gnome-shutdown", "Shutdown", "shutdown");
	menu_add_action_item (td->popup_menu, "gnome-suspend", "Suspend", "suspend");
	menu_add_action_item (td->popup_menu, "gnome-hibernate", "Hibernate", "hibernate");
#endif

#if GPM_HAS_QUIT
	menu_add_separator_item (td->popup_menu);
	item = gtk_image_menu_item_new_from_stock (GTK_STOCK_QUIT, NULL);
	g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK(callback_quit_activated), (gpointer) td->popup_menu);
	gtk_widget_show(item);
	gtk_menu_shell_append (GTK_MENU_SHELL (td->popup_menu), item);
#endif
}

/** Update icon by showing it, hiding it, or modifying it, as applicable
 *
 *  @param  tdicon		The icon pointer that needs to be updated
 */
void
update_icon (IconData *tdicon)
{
	g_assert (tdicon);
	g_debug ("update_icon");
	if (tdicon->tooltip)
		g_string_free (tdicon->tooltip, TRUE);
	tdicon->tooltip = NULL;

	if (tdicon->isMain) {
		tdicon->slotData = get_main_icon_slot ();
		if (tdicon->slotData)
			tdicon->tooltip = get_main_tooltip (&icon_main);
	} else {
		tdicon->slotData = get_low_icon_slot ();
		if (tdicon->slotData)
			tdicon->tooltip = get_low_tooltip (&icon_low);
	}

	if (tdicon->show && tdicon->slotData) {
		if (!tdicon->td) {
			icon_create (&(tdicon->td));
			if (tdicon->isMain && !(tdicon->td->popup_menu))
				menu_main_create (tdicon->td);
		}

		GdkPixbuf *pixbuf = create_icon_pixbuf (tdicon->slotData);
		icon_modify_image_pixbuf (tdicon->td, pixbuf);
		g_object_unref (pixbuf);

		if (tdicon->tooltip)
			icon_modify_tooltip (tdicon->td, tdicon->tooltip->str);
	} else {
		if (tdicon->td) {
			icon_destroy (&(tdicon->td));
			free_icon_structure (tdicon);
		}
	}
}
